package jalview.io;

import jalview.datamodel.HMMNode;
import jalview.datamodel.HiddenMarkovModel;
import jalview.datamodel.SequenceI;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.List;
import java.util.Scanner;


/**
 * reads in and writes out a HMMER standard file
 * 
 * 
 * @author TZVanaalten
 *
 */
public class HMMFile extends AlignFile
        implements AlignmentFileReaderI, AlignmentFileWriterI
{
  // HMM to store file data
  private HiddenMarkovModel hmm = new HiddenMarkovModel();




  // number of possible transitions
  private final int NUMBER_OF_TRANSITIONS = 7;

  private final String NEW_LINE = "\n";


  // file header
  String fileHeader;

  int numberOfSymbols;

  private final String SPACE = " ";

  private final String COMPO = "COMPO";

  private final String EMPTY = "";

  public HMMFile(FileParse source) throws IOException
  {
    super(false, source);
  }

  public HMMFile()
  {

  }

  public HiddenMarkovModel getHMM()
  {
    return hmm;
  }

  public void setHMM(HiddenMarkovModel model)
  {
    this.hmm = model;
  }

  public String getName()
  {
    return hmm.getName();
  }

  /**
   * reads data from HMM file
   * 
   * @throws IOException
   */
  @Override
  public void parse() throws IOException
  {
    parseFileProperties(dataIn);
    parseModel(dataIn);
  }



  /**
   * imports file properties from hmm file
   * 
   * @param input
   *          buffered reader used to read in file
   * @throws IOException
   */
  void parseFileProperties(BufferedReader input) throws IOException
  {
    boolean readingFile = true;
    fileHeader = input.readLine();
    String line = input.readLine();
    while (readingFile)
    {
      if (line != null)
      {
        Scanner parser = new Scanner(line);
        String next = parser.next();
        if ("HMM".equals(next)) // indicates start of HMM data (end of file
                              // properties)
        {
          readingFile = false;
          hmm.fillSymbols(parser);
          numberOfSymbols = hmm.getNumberOfSymbols();
        }
        else if ("STATS".equals(next))
        {
          parser.next();
          String key;
          String value;
          key = parser.next();
          value = parser.next() + SPACE + SPACE + parser.next();
          hmm.addFileProperty(key, value);
        }
        else
        {
          String key = next;
          String value = parser.next();
          while (parser.hasNext())
          {
            value = value + SPACE + parser.next();
          }
          hmm.addFileProperty(key, value);
        }
        parser.close();
      }
      line = input.readLine();
      if (line == null)
      {
        readingFile = false;
      }
    }

  }

  /**
   * parses the model data from the hmm file
   * 
   * @param input
   *          buffered reader used to read file
   * @throws IOException
   */
  void parseModel(BufferedReader input) throws IOException
  {
    for (int i = 0; i < hmm.getLength() + 1; i++)
    {
      hmm.getNodes().add(new HMMNode());
      String next;
      String line;
      line = input.readLine();
      Scanner matchReader = new Scanner(line);
      next = matchReader.next();
      if (next.equals(COMPO) || i > 0)
      {
        // stores match emission line in list
        List<Double> matches = new ArrayList<>();
        matches = fillList(matchReader, numberOfSymbols);
        hmm.getNodes().get(i).setMatchEmissions(matches);
        if (i > 0)
        {
          parseAnnotations(matchReader, i);
        }
      }
      matchReader.close();
      // stores insert emission line in list
      line = input.readLine();
      Scanner insertReader = new Scanner(line);
      List<Double> inserts = new ArrayList<>();
      inserts = fillList(insertReader, numberOfSymbols);
      hmm.getNodes().get(i).setInsertEmissions(inserts);
      insertReader.close();

      // stores state transition line in list
      line = input.readLine();
      Scanner transitionReader = new Scanner(line);
      List<Double> transitions = new ArrayList<>();
      transitions = fillList(transitionReader, NUMBER_OF_TRANSITIONS);
      hmm.getNodes().get(i).setStateTransitions(transitions);
      transitionReader.close();
    }

  }

  /**
   * parses annotations on match emission line
   * 
   * @param scanner
   *          scanner which is processing match emission line
   * @param index
   *          index of node which is beign scanned
   */
  void parseAnnotations(Scanner scanner, int index)
  {
    if (hmm.mapIsActive())
    {
      int column;
      column = scanner.nextInt();
      hmm.getNodes().get(index).setAlignmentColumn(column);
      hmm.getNodeLookup().put(column, index);
    }
    else
    {
      scanner.next();
    }

    char consensusR;
    consensusR = charValue(scanner.next());
    hmm.getNodes().get(index).setConsensusResidue(consensusR);

      char reference;
      reference = charValue(scanner.next());
      hmm.getNodes().get(index).setReferenceAnnotation(reference);


      char value;
      value = charValue(scanner.next());
      hmm.getNodes().get(index).setMaskValue(value);

    char consensusS;
    consensusS = charValue(scanner.next());
    hmm.getNodes().get(index).setConsensusStructure(consensusS);
  }


  /**
   * 
   * @param input
   *          scanner for line containing data to be transferred to list
   * @param numberOfElements
   *          number of elements in the list to be filled
   * @return filled list
   */
  static List<Double> fillList(Scanner input,
          int numberOfElements)
  {
    List<Double> list = new ArrayList<>();
    for (int i = 0; i < numberOfElements; i++)
    {

      String next = input.next();
      if (next.contains("*")) // state transitions to or from delete states
                              // occasionally have values of -infinity. These
                              // values are represented by an * in the .hmm
                              // file, and by a null value in the
                              // HiddenMarkovModel class
      {
        list.add(Double.NEGATIVE_INFINITY);
      }
      else
      {
        double prob = Double.valueOf(next);
        prob = Math.pow(Math.E, -prob);
        list.add(prob);
      }
    }
    return list;
  }

  
  /**
   * writes a HiddenMarkovModel to a file
   * 
   * @param exportLocation
   *          Filename, URL or Pasted String to write to
   * @throws FileNotFoundException
   * @throws UnsupportedEncodingException
   *
   **/
  
  public void exportFile(String exportLocation) throws IOException
  {
    StringBuilder file = new StringBuilder();
    appendFileProperties(file);
    appendModel(file);
    file.append("//");

    PrintWriter output = new PrintWriter(exportLocation);
    output.append(file);
    output.close();

  }

  String addData(int initialColumnSeparation,
          int columnSeparation, List<String> data)
  {
    String line = EMPTY;
    int index = 0;
    for (String value : data)
    {
      if (index == 0)
      {
        line += String.format("%" + initialColumnSeparation + "s", value);
      }
      else
      {
        line += String.format("%" + columnSeparation + "s", value);
      }
      index++;
    }
    return line;
  }

  List<String> charListToStringList(List<Character> list)
  {
    List<String> strList = new ArrayList<>();
    for (char value : list)
    {
      String strValue = Character.toString(value);
      strList.add(strValue);
    }
    return strList;
  }

  List<String> doubleListToStringList(List<Double> list,
          int noOfDecimals)
  {
    List<String> strList = new ArrayList<>();
    for (double value : list)
    {
      String strValue;
      if (value > 0)
      {
        strValue = String.format("%.5f", value);

      }
      else if (value == -0.00000d)
      {
        strValue = "0.00000";
      }
      else
      {
        strValue = "*";
      }

      strList.add(strValue);
    }
    return strList;
  }

  List<String> stringArrayToStringList(String[] array)
  {
    List<String> list = new ArrayList<>();
    for (String value : array)
    {
      list.add(value);
    }

    return list;
  }

  void appendModel(StringBuilder file)
  {
    String symbolLine = "HMM";
    List<Character> charSymbols = hmm.getSymbols();
    List<String> strSymbols;
    strSymbols = charListToStringList(charSymbols);
    symbolLine += addData(11, 9, strSymbols);
    file.append(symbolLine + NEW_LINE);

    String transitionTypeLine = "";
    List<String> transitionTypes;
    transitionTypes = stringArrayToStringList(hmm.getTransitionTypes());
    transitionTypeLine += addData(16, 9, transitionTypes);
    file.append(transitionTypeLine + NEW_LINE);

    int length = hmm.getLength();

    for (int node = 0; node <= length; node++)
    {
      String matchLine;
      if (node == 0)
      {
        matchLine = String.format("%7s", "COMPO");
      }
      else
      {
        matchLine = String.format("%7s", node);
      }

      List<String> strMatches;
      List<Double> doubleMatches;
      doubleMatches = hmm.getNode(node).getMatchEmissions();
      convertListToLogSpace(doubleMatches);
      strMatches = doubleListToStringList(doubleMatches, 5);
      matchLine += addData(10, 9, strMatches);


      if (node != 0)
      {
        matchLine += SPACE + hmm.getNodeAlignmentColumn(node);
        matchLine += SPACE + hmm.getConsensusResidue(node);
        matchLine += SPACE + hmm.getReferenceAnnotation(node);
        matchLine += SPACE + hmm.getMaskedValue(node);
        matchLine += SPACE + hmm.getConsensusStructure(node);

      }

      file.append(matchLine + NEW_LINE);
      
      String insertLine = EMPTY;
      List<String> strInserts;
      List<Double> doubleInserts;
      doubleInserts = hmm.getNode(node).getInsertEmissions();
      convertListToLogSpace(doubleInserts);
      strInserts = doubleListToStringList(doubleInserts, 5);
      insertLine += addData(17, 9, strInserts);

      file.append(insertLine + NEW_LINE);

      String transitionLine = EMPTY;
      List<String> strTransitions;
      List<Double> doubleTransitions;
      doubleTransitions = hmm.getNode(node).getStateTransitions();
      convertListToLogSpace(doubleTransitions);
      strTransitions = doubleListToStringList(doubleTransitions, 5);
      transitionLine += addData(17, 9, strTransitions);

      file.append(transitionLine + NEW_LINE);
    }
  }

  void appendFileProperties(StringBuilder file)
  {
    String line;

    file.append(fileHeader + NEW_LINE);
    
    line = String.format("%-5s %1s", "NAME", hmm.getName());
    file.append((line + NEW_LINE));

    if (hmm.getAccessionNumber() != null)
    {
    line = String.format("%-5s %1s", "ACC", hmm.getAccessionNumber());
    file.append((line + NEW_LINE));
    }

    if (hmm.getDescription() != null)
    {
    line = String.format("%-5s %1s", "DESC", hmm.getDescription());
    file.append((line + NEW_LINE));
    }
    line = String.format("%-5s %1s", "LENG", hmm.getLength());
    file.append((line + NEW_LINE));

    if (hmm.getMaxInstanceLength() != null)
    {
    line = String.format("%-5s %1s", "MAXL", hmm.getMaxInstanceLength());
    file.append((line + NEW_LINE));
    }
    line = String.format("%-5s %1s", "ALPH", hmm.getAlphabetType());
    file.append((line + NEW_LINE));

    boolean status;
    String statusStr;

    status = hmm.referenceAnnotationIsActive();
    statusStr = HiddenMarkovModel.findStringFromBoolean(status);
    line = String.format("%-5s %1s", "RF",
            statusStr);
    file.append((line + NEW_LINE));

    status = hmm.maskValueIsActive();
    statusStr = HiddenMarkovModel.findStringFromBoolean(status);
    line = String.format("%-5s %1s", "MM",
            statusStr);
    file.append((line + NEW_LINE));
    
    status = hmm.consensusResidueIsActive();
    statusStr = HiddenMarkovModel.findStringFromBoolean(status);
    line = String.format("%-5s %1s", "CONS",
            statusStr);
    file.append((line + NEW_LINE));

    status = hmm.consensusStructureIsActive();
    statusStr = HiddenMarkovModel.findStringFromBoolean(status);
    line = String.format("%-5s %1s", "CS",
            statusStr);
    file.append((line + NEW_LINE));

    status = hmm.mapIsActive();
    statusStr = HiddenMarkovModel.findStringFromBoolean(status);
    line = String.format("%-5s %1s", "MAP",
            statusStr);
    file.append((line + NEW_LINE));


    if (hmm.getDate() != null)
    {
    line = String.format("%-5s %1s", "DATE", hmm.getDate());
    file.append((line + NEW_LINE));
    }
    if (hmm.getNumberOfSequences() != null)
    {
    line = String.format("%-5s %1s", "NSEQ", hmm.getNumberOfSequences());
    file.append((line + NEW_LINE));
    }
    if (hmm.getEffectiveNumberOfSequences() != null)
    {
    line = String.format("%-5s %1s", "EFFN",
            hmm.getEffectiveNumberOfSequences());
    file.append((line + NEW_LINE));
    }
    if (hmm.getCheckSum() != null)
    {
    line = String.format("%-5s %1s", "CKSUM", hmm.getCheckSum());
    file.append((line + NEW_LINE));
    }
    if (hmm.getGatheringThreshold() != null)
    {
    line = String.format("%-5s %1s", "GA", hmm.getGatheringThreshold());
    file.append((line + NEW_LINE));
    }

    if (hmm.getTrustedCutoff() != null)
    {
    line = String.format("%-5s %1s", "TC", hmm.getTrustedCutoff());
    file.append((line + NEW_LINE));
    }
    if (hmm.getNoiseCutoff() != null)
    {
    line = String.format("%-5s %1s", "NC", hmm.getNoiseCutoff());
    file.append((line + NEW_LINE));
    }
    if (hmm.getMSV() != null)
    {
      line = String.format("%-19s %18s", "STATS LOCAL MSV", hmm.getMSV());
      file.append((line + NEW_LINE));

      line = String.format("%-19s %18s", "STATS LOCAL VITERBI",
              hmm.getViterbi());
      file.append((line + NEW_LINE));
    
      line = String.format("%-19s %18s", "STATS LOCAL FORWARD",
              hmm.getForward());
      file.append((line + NEW_LINE));
    }
  }



  char charValue(String string)
  {
    char character;
    character = string.charAt(0);
    return character;
  }

  @Override
  public String print(SequenceI[] seqs, boolean jvsuffix)
  {

    return null;
  }

  void convertListToLogSpace(List<Double> list)
  {

    for (int i = 0; i < list.size(); i++)
    {
      double prob = list.get(i);
      double logProb = -1 * Math.log(prob);

      list.set(i, logProb);
    }


  }
}

